if $0 == __FILE__
require 'rake4latex'
end

module Rake4LaTeX
=begin rdoc
This class collects informations for an existing (La)TeX-File.

It can be used to determine dependecies of a LaTeX project.

What's not done:
* check kpsewhich if files are not found
* Analyse styles
* 

=end
class TeXfile

  #fixme: global logger
  @@log = Log4r::Logger.new("LaTeXRunner", Log4r::ERROR)
  @@log.outputters = Log4r::StdoutOutputter.new('log_stdout')

=begin rdoc
Define a TeXfile.

If the include-analyse should work correct,
the work directory should be the place, where you find your file.

If filename is :dummy, you get a dummy-file for little extracts.
=end
  def initialize( filename )

    @filename = filename      
    @@log.debug("TeXfile to analyse: #{@filename}")
    #~ @@log.warn("Start TeXfile analyse in foreign directory") unless File.dirname(filename) == '.'

    #List of filenames of the includes
    @includes = []
    #Hash with the objects for the include files.
    @include_object = {}
    @bibliography = []
    @gloss = {} #bib-files for gloss.sty
    @splitindex = {}
    
    if @filename != :dummy
      @basename = File.basename(filename)
      raise ArgumentError, "File #{@filename} not found" unless File.exist?(@filename)
      analyse( File.read(@filename) )
    end
  end
=begin rdoc
List of all includes.

Options:
* :recurssive - get all includes, including includes of includes
* :uniq - delete double includes
=end
  def includes( *options )
    if options.include?(:recursive)
      includes = tree.flatten - [ @filename ]
    else
      includes = @includes
    end
    includes.uniq! if options.include?(:uniq)
    includes
  end
=begin rdoc
If bibTeX is used, this Array contains all bibliography.
=end
  attr_reader :bibliography
=begin rdoc
If gloss.sty is used, this Hash contains all glossaries and there bibliography.
=end
  attr_reader :gloss
=begin rdoc
If splitindex is used, this Array contains all indeces.
=end
  attr_reader :splitindex
  
=begin rdoc
Analyse the file.

Search dependecies by sanning for 
* \input/\include
* \bibliography
* \printgloss (gloss.sty, like bibliography)
* \newindex (splitindex, like makeindex)
=end
  def analyse(content )

    content.each_line{ |line|
      #Skip comments
      #Remarks/Bug: 
      #* \% is not detected.
      #* \iffalse...\fi is not detected.
      line.sub!(/%.*/, '')
      break if line =~ %r{\\endinput(\s|\Z)}  #Stop at \endinput
      #Detect \macro[]{}
      line.scan(/\\(include|input|includeonly|bibliography|printgloss|newindex)(\[.*?\])?\{([^}]+)\}/){|m|
        case m[0]
          #~ when 'includeonly'  
          when 'include', 'input'
            includename = m[2].ext('tex')  #add the extension tex if not already done.
            @includes << includename
          when 'bibliography'
            #~ @bibliography << m[2].ext('bib')
            @bibliography << m[2].split(/,/).map{|f| f.ext('bib')}
            @bibliography.flatten!
          when 'printgloss'
            m[1]  = 'gls' unless m[1]
            @gloss[m[1]] ||= [] << m[2].split(/,/).map{|f| f.ext('bib')}
            @gloss[m[1]].flatten!
          when 'newindex'
            #m[1] contains the optional name
            @splitindex[m[2]] = m[1]
        end #case m[0]
      }#line.scan
    }#File.foreach

    @includes.each{|includename|
      if File.exist?(includename)
        if @include_object[includename]
          #fixme recursiv...
          #This check detects only a double input in one file.
          @@log.warn("Double input of #{includename} in #{@filename}")
        else
          @include_object[includename] = TeXfile.new(includename)
        end
      else
        @@log.warn("Include-File #{includename} not found")
      end
    }

    #Check the bibTeX-File in the TeXMF-trees (using kpsewhich)
    @bibliography.map!{|bib| kpsewhich(bib) }
    @gloss.each{|key,bibs| bibs.map!{|bib| kpsewhich(bib) }}
  end
=begin rdoc
Check with kpsewhich to get a file location.

If nothing is found, take the original name.
=end
  def kpsewhich( filename )
    hit = `kpsewhich #{filename}`
    @@log.warn("Include-File #{filename} not found") if hit.empty?
    hit.empty? ? filename : hit.strip
  end
=begin
Get a tree-like view of all includes.
=end
  def tree()
    tree = [@filename]
    #Only first occurence and without a sequence
    #~ @include_object.each{|name, inc|
      #~ tree << inc.tree
    #~ }

    @includes.each{|name|
      if @include_object[name]
        tree << @include_object[name].tree
      else
        tree << name
      end
    }

    tree
  end
  def inspect()
    "<TeXfile #{@filename}>"
  end
end #TeXfile


#List of all options for Rake4LaTeX.build_rakefile
BUILD_RAKEFILE_OPTIONS = [
    :pdf, #add rule to create a pdf
    :dvi,  #add rule to create a dvi
    :dependecies, #Analyse dependecies and add them
    :bib, #Add bibliography-dependecies (1)
    :gloss, #Add gloss.sty-dependecies
    :splitindex
]
=begin rdoc
Build a rakefile

Options:
* :pdf - add rule to create a pdf
* :dvi - add rule to create a dvi
* :dependecies - Analyse dependecies and add them
* :bib: Add bibliography-dependecies (1)
* :gloss: Add gloss.sty-dependecies
* :splitindex
* :all - all options,

Remarks:

(1) The bibTeX-files are searched via kpsewhich.
local files get a "./" in front of the name.
=end
def self.build_rakefile( filename, *options )
  
  options.concat(BUILD_RAKEFILE_OPTIONS) if options.delete(:all )
  ( options - BUILD_RAKEFILE_OPTIONS ).each{|opt|
    Rake4LaTeX::Logger.warn("Undefined option #{opt} in Rake4LaTeX.build_rakefile")
  }
  basename = filename.ext()
  tex_name = filename.ext('tex')
  pdf_name = filename.ext('pdf')
  dvi_name = filename.ext('dvi')
  aux_name = filename.ext('aux')
  texfile = Rake4LaTeX::TeXfile.new(tex_name)
  #
  #Template for a basic rakefile
  #
  template = <<Template
=begin
Rakefile for LaTeX-project #{basename}

Generated by Rake4LaTeX.build_rakefile, #{Date.today}
Options: #{options.inspect}

=end
require 'rake4latex'

#Change default
#~ Rake4LaTeX::DEFAULT_SETTINGS[:texerrors_allowed] = true

#Needed for some actions without a previous TeX-call
task :basefile => '#{tex_name}'

Template

  template << <<Template
#
# Define Dependecies
#
Template

  template << "file '#{pdf_name}' => '#{tex_name}'\n" if options.include?(:pdf)
  template << "file '#{dvi_name}' => '#{tex_name}'\n" if options.include?(:dvi)
  if options.include?(:dependecies) 
    includes = texfile.includes(:uniq, :recursive)
    if ! includes.empty?
      template << <<Template
#{texfile.tree.to_yaml.gsub(/^/, "# ")}
dependecies = [ 
#{includes.map{|f| "  \"#{f}\",\n"}.join}]
Template
      template << "file '#{pdf_name}' => dependecies\n" if options.include?(:pdf)
      template << "file '#{dvi_name}' => dependecies\n" if options.include?(:dvi)
    end
  end #options.include?(:dependecies)

  if options.include?(:bib) and ! texfile.bibliography.empty?
    template << <<Template

=begin
bib-file dependecies
=end
Template
    template << "file '#{basename}.bbl' => #{texfile.bibliography.uniq.inspect}\n"
    template << "file '#{pdf_name}' => #{texfile.bibliography.uniq.inspect}\n" if options.include?(:pdf)
    template << "file '#{dvi_name}' => #{texfile.bibliography.uniq.inspect}\n" if options.include?(:dvi)
  end #options.include?(:bib)
  
  if options.include?(:gloss) and ! texfile.gloss.empty?
    template << <<Template

=begin
bib-file dependecies for gloss-style
=end
Template
    texfile.gloss.each{|glossar, bib|
      template << "# glossar #{glossar}\n"
      template << "file '#{basename}.#{glossar}.bbl' => #{bib.uniq.inspect}\n"
      template << "file '#{basename}.#{glossar}.bbl' => '#{aux_name}'\n"
      
      template << "file '#{pdf_name}' => #{bib.uniq.inspect}\n" if options.include?(:pdf)
      template << "file '#{dvi_name}' => #{bib.uniq.inspect}\n" if options.include?(:dvi)
    }
  end #options.include?(:bib)

  if options.include?(:splitindex) and ! texfile.splitindex.empty?
    template << <<Template

=begin
dependecies for splitindex
=end
Template
    texfile.splitindex.each{|splitindex, indexname|
      template << "# Index #{splitindex} #{indexname}\n"
      template << "file '#{basename}-#{splitindex}.ind' => '#{basename}-#{splitindex}.idx'\n"
      template << "file '#{pdf_name}' => '#{basename}-#{splitindex}.ind'\n" if options.include?(:pdf)
      template << "file '#{dvi_name}' => '#{basename}-#{splitindex}.ind'\n" if options.include?(:dvi)
    }
  end #options.include?(:bib)


  template << <<Template

#Force the compilation
task :touch => '#{tex_name}'

Template

  template << <<Template
=begin
Define the default tasks
=end
#~ task :default => :touch  #Force the first TeX-call
Template

  template << "task :default => '#{pdf_name}' #build the pdf\n" if options.include?(:pdf)
  template << "task :default => '#{dvi_name}' #build the dvi\n" if options.include?(:dvi)

  template << <<Template

=begin
Closing Tasks
=end
task :default => :statistic #Get an overview on errors and warnings
#~ task :default => :log_overview #Show log overview-file
#~ task :default => :log_overview_file #Build a log overview-file
task :default => :clean #delete helpfiles

Template

  template << <<Template
=begin
Make the rakefile self-executable
=end
if $0 == __FILE__
  app = Rake.application
  app[:default].invoke  #start default tasks
end
Template

  template
end #self.template( filename = nil)


end #Rake4LaTeX

if $0 == __FILE__
  #fixmes:
  #Parameter "xx.sty" -> analyse also this style

  #~ file = Rake4LaTeX::TeXfile.new(:dummy)
  #~ file.analyse(<<'xx'
#~ \input{aaa}
#~ \bibliography{test}
#~ \bibliography{testdocument}
#~ xx
#~ )
  #~ puts file
  #~ puts file.includes.inspect
  #~ puts file.bibliography.inspect
  #~ puts file.splitindex.inspect
  #~ puts file.tree.to_yaml
  #~ exit

Dir.chdir('../../test'){
  #~ Dir.chdir('gloss'){
  Dir.chdir('bibtex'){
  #~ Dir.chdir('includes'){
  #~ Dir.chdir('splitindex'){
    #~ file = Rake4LaTeX::TeXfile.new('testdocument.tex')
    #~ puts file
    #~ puts file.includes.inspect
    #~ puts file.bibliography.inspect
    #~ puts file.splitindex.inspect
    #~ puts file.tree.to_yaml
    
    #~ puts Rake4LaTeX.build_rakefile('testdocument.tex', :dvi, :pdf, :dependecies, :bib, :gloss, :splitindex )
    puts Rake4LaTeX.build_rakefile('testdocument.tex', :all)
  }
}

end